# Implementations of LS-NMF
# 
# Reference:
# LS-NMF: a modified non-negative matrix factorization algorithm utilizing uncertainty estimates.
# ﻿Wang, Guoli, Andrew V Kossenkov, and Michael F Ochs. 
# BMC bioinformatics 7 (January 2006): 175. http://www.ncbi.nlm.nih.gov/pubmed/16569230.
#  ﻿
# Author: Renaud Gaujoux
# Creation: 09 Nov 2011
###############################################################################


#' Multiplicative Updates for LS-NMF
#' 
#'  
#' 
#' @param i current iteration
#' @param X target matrix
#' @param object current NMF model
#' @param sigma uncertainty estimates, used to weight each entry in X 
#' @param eps 
#' @param ... 
#'  
#' @return 
#' @author Renaud Gaujoux
#' @name LSNMF-NMF
nmf.update.lsnmf <- function(i, X, object, weight, eps=10^-9, ...)
{
	if( i == 1 ){# pre-compute weighted target matrix
		staticVar('wX', X * weight, init=TRUE)
	}
	
	# retrieve weighted target matrix
	wX <- staticVar('wX')
	
	# retrieve each factor
	w <- basis(object); h <- coef(object);	
	
	# compute the estimate WH
	wh <- fitted(object) * weight
	
	# euclidean-reducing NMF iterations	
	# H_au = H_au (W^T V/sigma)_au / (W^T (W H)/sigma)_au
	h <- R_std.euclidean.update.h(wX, w, h, wh=wh, eps=eps)	
	
	# update H and recompute the estimate WH
	coef(object) <- h;
	wh <- fitted(object) * weight
	
	# W_ia = W_ia (V/sigma H^T)_ia / ((W H)/sigma H^T)_ia and columns are rescaled after each iteration	
	w <- R_std.euclidean.update.w(wX, w, h, wh=wh, eps=eps)	
	
	#return the modified data
	basis(object) <- w	
	return(object)
}

#' Weigthed Residual Sum of Squares
wrss <- function(X, object, weight){
	sum( ((X - fitted(object)) * weight)^2 )/2
}

#' Registration of LS-NMF
.nmf.plugin.lsnmf <- function(){
	list(
	new('NMFStrategyIterative', name='ls-nmf', objective=wrss
			, Update=nmf.update.lsnmf
			, Stop='nmf.stop.stationary')
	)
}

#' Unit test for the LS-NMF algorithm
runit.lsnmf <- function(){
	
	set.seed(12345)
	X <- rmatrix(100,20)
	
	res <- nmf(X, 3, 'ls-nmf', weight=1, seed=1)	
	res2 <- nmf(X, 3, '.R#lee', rescale=FALSE, seed=1, .stop=nmf.stop.stationary)
	tol <- 10^-14
	checkTrue( nmf.equal(res, res2, identical=FALSE, tol=tol ), paste("LS-NMF with weight = 1 and .R#Lee (no scale + stationary) give identical results at tolerance=", tol))	
	
}
